package main

import (
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"os"
	"path/filepath"
	"runtime"
	"sync"
	"time"

	flag "github.com/spf13/pflag"
)

// Fdhashes struct for holding all informations about one folder.
type Fdhashes struct {
	Path   string
	Hashes map[string]string
	Times  map[string]time.Time
}

var hashes map[string]Fdhashes

var wg sync.WaitGroup
var mu sync.RWMutex
var driveLetter string

var rewrite bool

func init() {
	flag.BoolVarP(&rewrite, "rewrite", "r", false, "rewrite all fhhashes files.")
}

func main() {
	log.Println("starting GoHash")
	runtime.GOMAXPROCS(5)
	hashes = make(map[string]Fdhashes)
	flag.Parse()
	myFile := flag.Arg(0)
	file, err := os.Stat(myFile)
	if os.IsNotExist(err) {
		log.Fatalln("File does not exists:", myFile)
	}
	if file.IsDir() {
		log.Println("start with folder:", myFile)
		driveLetter = ""
		if runtime.GOOS == "windows" {
			driveLetter = filepath.VolumeName(myFile) + "/"
		}
		processFolder(myFile)
	} else {
		log.Printf("file %s has hash %s\n", myFile, getSha256Hash(myFile))
	}

	fmt.Println("waiting")
	wg.Wait()
	log.Println("done")
}

func getSha256Hash(fileStr string) string {
	f, err := os.Open(fileStr)
	if err != nil {
		log.Fatal(err)
	}
	defer f.Close()

	h := sha256.New()
	if _, err := io.Copy(h, f); err != nil {
		log.Fatal(err)
	}
	return hex.EncodeToString(h.Sum(nil))
}

func outputHash(fileStr string) {
	var hashFile Fdhashes
	doHash := true
	defer wg.Done()
	dir, fileName := filepath.Split(fileStr)
	if fileName == ".fdhashes3" {
		return
	}
	// checking if hash is present
	mu.Lock()
	hashFile, ok := hashes[dir]
	if !ok {
		_, err := os.Stat(dir + ".fdhashes3")
		if os.IsNotExist(err) {
			hashFile = Fdhashes{Path: dir, Hashes: make(map[string]string), Times: make(map[string]time.Time)}
		} else {
			hashFile = loadHashfile(dir + ".fdhashes3")
		}
		hashes[dir] = hashFile
		saveHashfile(hashFile)
	}
	_, ok = hashFile.Hashes[fileName]
	mu.Unlock()
	doHash = !ok
	// checking if dattime is identically
	file, _ := os.Stat(fileStr)
	time := file.ModTime()
	savedTime, ok := hashFile.Times[fileName]
	if !time.Equal(savedTime) || !ok {
		doHash = true
	}
	if doHash {
		hash := getSha256Hash(fileStr)
		mu.Lock()
		hashFile.Hashes[fileName] = hash
		hashFile.Times[fileName] = time
		hashes[dir] = hashFile
		saveHashfile(hashFile)
		mu.Unlock()
		log.Printf("file \"%s\" has hash \"%s\"\n", fileStr, hash)
	}
}

var count int
var addWork int

func processFolder(folder string) {
	count = 0
	addWork = 0
	err := filepath.Walk(folder, func(path string, info os.FileInfo, err error) error {
		count++
		if (count % 100) == 0 {
			fmt.Print(".")
		}
		if (count % 10000) == 0 {
			fmt.Println()
		}
		filename := info.Name()
		if filename[0:1] != "." {
			if info.IsDir() {
				fmt.Println(path)
			}
			if !info.IsDir() {
				addWork++
				wg.Add(1)
				go outputHash(path)
				if addWork > 100 {
					fmt.Println("x")
					wg.Wait()
					addWork = 0
				}
			}
		}
		return nil
	})
	if err != nil {
		panic(err)
	}
}

func saveHashfile(hashFile Fdhashes) {
	b, err := json.Marshal(hashFile)
	if err != nil {
		fmt.Println(err)
		return
	}
	err = ioutil.WriteFile(hashFile.Path+".fdhashes3", b, 0644)
	if err != nil {
		panic(err)
	}
}

func loadHashfile(fileStr string) Fdhashes {
	dir, _ := filepath.Split(fileStr)
	data := Fdhashes{Path: dir, Hashes: make(map[string]string), Times: make(map[string]time.Time)}
	if !rewrite {
		file, err := ioutil.ReadFile(fileStr)
		if err != nil {
			panic(err)
		}
		err = json.Unmarshal([]byte(file), &data)
		if err != nil {
			log.Printf("can't read file %s", fileStr)
		}
	}
	if data.Path == "" {
		data.Path = dir
	}
	return data
}
